from typing import Any, List, Optional

from pydantic import BaseModel, Field

from .base import CaseInsensitiveStrEnum
from .data_types import ClickhouseDataType


class TableMapping(BaseModel):
    source_id: str
    field_name: str
    column_name: str
    column_type: ClickhouseDataType


class SinkType(CaseInsensitiveStrEnum):
    CLICKHOUSE = "clickhouse"


class SinkConfig(BaseModel):
    type: SinkType = SinkType.CLICKHOUSE
    provider: Optional[str] = Field(default=None)
    host: str
    port: str
    http_port: Optional[str] = Field(default=None)
    database: str
    username: str
    password: str
    secure: bool = Field(default=False)
    skip_certificate_verification: bool = Field(default=False)
    max_batch_size: int = Field(default=1000)
    max_delay_time: str = Field(default="10m")
    table: str
    table_mapping: List[TableMapping]

    def update(self, patch: "SinkConfigPatch") -> "SinkConfig":
        """Apply a patch to this sink config."""
        update_dict: dict[str, Any] = {}

        # Check each field explicitly to handle model instances properly
        if patch.provider is not None:
            update_dict["provider"] = patch.provider
        if patch.host is not None:
            update_dict["host"] = patch.host
        if patch.port is not None:
            update_dict["port"] = patch.port
        if patch.http_port is not None:
            update_dict["http_port"] = patch.http_port
        if patch.database is not None:
            update_dict["database"] = patch.database
        if patch.username is not None:
            update_dict["username"] = patch.username
        if patch.password is not None:
            update_dict["password"] = patch.password
        if patch.secure is not None:
            update_dict["secure"] = patch.secure
        if patch.skip_certificate_verification is not None:
            update_dict["skip_certificate_verification"] = (
                patch.skip_certificate_verification
            )
        if patch.max_batch_size is not None:
            update_dict["max_batch_size"] = patch.max_batch_size
        if patch.max_delay_time is not None:
            update_dict["max_delay_time"] = patch.max_delay_time
        if patch.table is not None:
            update_dict["table"] = patch.table
        if patch.table_mapping is not None:
            update_dict["table_mapping"] = patch.table_mapping

        if update_dict:
            return self.model_copy(update=update_dict)
        return self


class SinkConfigPatch(BaseModel):
    provider: Optional[str] = Field(default=None)
    host: Optional[str] = Field(default=None)
    port: Optional[str] = Field(default=None)
    http_port: Optional[str] = Field(default=None)
    database: Optional[str] = Field(default=None)
    username: Optional[str] = Field(default=None)
    password: Optional[str] = Field(default=None)
    secure: Optional[bool] = Field(default=None)
    skip_certificate_verification: Optional[bool] = Field(default=None)
    max_batch_size: Optional[int] = Field(default=None)
    max_delay_time: Optional[str] = Field(default=None)
    table: Optional[str] = Field(default=None)
    table_mapping: Optional[List[TableMapping]] = Field(default=None)



from enum import Enum


class CaseInsensitiveStrEnum(str, Enum):
    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(f"Invalid value: {value}")

    def __str__(self):
        return str(self.value)



import warnings
from typing import Any, List, Optional

from pydantic import BaseModel, Field, ValidationInfo, field_validator, model_validator

from .base import CaseInsensitiveStrEnum
from .data_types import KafkaDataType


class KafkaProtocol(CaseInsensitiveStrEnum):
    SSL = "SSL"
    SASL_SSL = "SASL_SSL"
    SASL_PLAINTEXT = "SASL_PLAINTEXT"
    PLAINTEXT = "PLAINTEXT"


class KafkaMechanism(CaseInsensitiveStrEnum):
    SCRAM_SHA_256 = "SCRAM-SHA-256"
    SCRAM_SHA_512 = "SCRAM-SHA-512"
    PLAIN = "PLAIN"
    GSSAPI = "GSSAPI"
    NO_AUTH = "NO_AUTH"


class SchemaField(BaseModel):
    name: str
    type: KafkaDataType


class SchemaType(CaseInsensitiveStrEnum):
    JSON = "json"


class Schema(BaseModel):
    type: SchemaType = SchemaType.JSON
    fields: List[SchemaField]


class DeduplicationConfig(BaseModel):
    enabled: bool = False
    id_field: Optional[str] = Field(default=None)
    id_field_type: Optional[KafkaDataType] = Field(default=None)
    time_window: Optional[str] = Field(default=None)

    def update(self, patch: "DeduplicationConfigPatch") -> "DeduplicationConfig":
        """Apply a patch to this deduplication config."""
        update_dict: dict[str, Any] = {}

        # Check each field explicitly - use model_fields_set to distinguish
        # between "not provided" and "set to None"
        fields_set = (
            patch.model_fields_set if hasattr(patch, "model_fields_set") else set()
        )

        if "enabled" in fields_set or patch.enabled is not None:
            update_dict["enabled"] = patch.enabled
        if "id_field" in fields_set:
            update_dict["id_field"] = patch.id_field
        if "id_field_type" in fields_set:
            update_dict["id_field_type"] = patch.id_field_type
        if "time_window" in fields_set:
            update_dict["time_window"] = patch.time_window

        if update_dict:
            return self.model_copy(update=update_dict)
        return self

    @model_validator(mode="before")
    @classmethod
    def validate_deduplication_fields(cls, values):
        """Validate deduplication fields based on enabled status."""
        if isinstance(values, dict):
            enabled = values.get("enabled", False)

            # If deduplication is disabled, allow empty strings
            if not enabled:
                # Convert empty strings to None for enum fields
                if values.get("id_field_type") == "":
                    values["id_field_type"] = None
                if values.get("id_field") == "":
                    values["id_field"] = None
                if values.get("time_window") == "":
                    values["time_window"] = None
            else:
                # If enabled, ensure required fields are present and not empty
                for field_name in ["id_field", "id_field_type", "time_window"]:
                    field_value = values.get(field_name)
                    if field_value is None or field_value == "":
                        raise ValueError(
                            f"{field_name} is required when deduplication is enabled"
                        )

                # Validate id_field_type is a valid type when enabled
                id_field_type = values.get("id_field_type")
                if id_field_type not in [KafkaDataType.STRING, KafkaDataType.INT]:
                    raise ValueError(
                        "id_field_type must be a string or int when "
                        "deduplication is enabled"
                    )

        return values


class ConsumerGroupOffset(CaseInsensitiveStrEnum):
    LATEST = "latest"
    EARLIEST = "earliest"


class TopicConfig(BaseModel):
    consumer_group_initial_offset: ConsumerGroupOffset = ConsumerGroupOffset.LATEST
    name: str
    event_schema: Schema = Field(alias="schema")
    deduplication: Optional[DeduplicationConfig] = Field(default=DeduplicationConfig())
    replicas: Optional[int] = Field(default=1)

    @field_validator("deduplication")
    @classmethod
    def validate_deduplication_id_field(
        cls, v: DeduplicationConfig, info: ValidationInfo
    ) -> DeduplicationConfig:
        """
        Validate that the deduplication ID field exists in the
        schema and has matching type.
        """
        if v is None or not v.enabled:
            return v

        # Skip validation if id_field is empty when deduplication is disabled
        if not v.id_field or v.id_field == "":
            return v

        # Get the schema from the parent model's data
        schema = info.data.get("event_schema", {})
        if isinstance(schema, dict):
            fields = schema.get("fields", [])
        else:
            fields = schema.fields

        # Find the field in the schema
        field = next((f for f in fields if f.name == v.id_field), None)
        if not field:
            raise ValueError(
                f"Deduplication ID field '{v.id_field}' does not exist in "
                "the event schema"
            )

        return v

    @field_validator("replicas")
    @classmethod
    def validate_replicas(cls, v: int) -> int:
        if v < 1:
            raise ValueError("Replicas must be at least 1")
        return v


class KafkaConnectionParams(BaseModel):
    brokers: List[str]
    protocol: KafkaProtocol
    mechanism: Optional[KafkaMechanism] = Field(default=None)
    username: Optional[str] = Field(default=None)
    password: Optional[str] = Field(default=None)
    root_ca: Optional[str] = Field(default=None)
    kerberos_service_name: Optional[str] = Field(default=None)
    kerberos_keytab: Optional[str] = Field(default=None)
    kerberos_realm: Optional[str] = Field(default=None)
    kerberos_config: Optional[str] = Field(default=None)
    skip_auth: bool = Field(
        default=False,
        exclude=True,
        description='skip_auth is deprecated, use mechanism "NO_AUTH" instead or \
        set skip_tls_verification to True if you want to skip TLS verification',
    )
    skip_tls_verification: bool = Field(default=False)

    @model_validator(mode="before")
    def empty_str_to_none(values):
        if values.get("mechanism", None) == "":
            values["mechanism"] = None
        return values

    def update(self, patch: "KafkaConnectionParamsPatch") -> "KafkaConnectionParams":
        """Apply a patch to this connection params config."""
        current_dict = self.model_dump()
        patch_dict = patch.model_dump(exclude_none=True)
        merged_dict = {**current_dict, **patch_dict}
        return KafkaConnectionParams.model_validate(merged_dict)

    @model_validator(mode="after")
    def correct_skip_auth(self) -> "KafkaConnectionParams":
        """
        While the skip_auth field is deprecated,
        we need to make sure the mechanism is NO_AUTH if skip_auth is True
        """
        if self.skip_auth:
            if self.mechanism is not None and self.mechanism != KafkaMechanism.NO_AUTH:
                warnings.warn(
                    f"Mechanism is set to {self.mechanism}, but skip_auth is True. \
                    Setting mechanism to NO_AUTH.",
                    category=RuntimeWarning,
                    stacklevel=1,
                )
            self.mechanism = KafkaMechanism.NO_AUTH
        return self


class SourceType(CaseInsensitiveStrEnum):
    KAFKA = "kafka"


class SourceConfig(BaseModel):
    type: SourceType = SourceType.KAFKA
    provider: Optional[str] = Field(default=None)
    connection_params: KafkaConnectionParams
    topics: List[TopicConfig]

    def update(self, patch: "SourceConfigPatch") -> "SourceConfig":
        """Apply a patch to this source config."""
        update_dict: dict[str, Any] = {}

        if patch.type is not None:
            update_dict["type"] = patch.type
        if patch.provider is not None:
            update_dict["provider"] = patch.provider

        # Handle connection_params patch
        if patch.connection_params is not None:
            update_dict["connection_params"] = self.connection_params.update(
                patch.connection_params
            )

        # Handle topics patch - full replacement only if provided
        if patch.topics is not None:
            update_dict["topics"] = patch.topics

        if update_dict:
            return self.model_copy(update=update_dict)
        return self


class DeduplicationConfigPatch(BaseModel):
    enabled: Optional[bool] = Field(default=None)
    id_field: Optional[str] = Field(default=None)
    id_field_type: Optional[KafkaDataType] = Field(default=None)
    time_window: Optional[str] = Field(default=None)


class KafkaConnectionParamsPatch(BaseModel):
    brokers: Optional[List[str]] = Field(default=None)
    protocol: Optional[KafkaProtocol] = Field(default=None)
    mechanism: Optional[KafkaMechanism] = Field(default=None)
    username: Optional[str] = Field(default=None)
    password: Optional[str] = Field(default=None)
    root_ca: Optional[str] = Field(default=None)
    kerberos_service_name: Optional[str] = Field(default=None)
    kerberos_keytab: Optional[str] = Field(default=None)
    kerberos_realm: Optional[str] = Field(default=None)
    kerberos_config: Optional[str] = Field(default=None)
    skip_auth: Optional[bool] = Field(default=None)


class SourceConfigPatch(BaseModel):
    type: Optional[SourceType] = Field(default=None)
    provider: Optional[str] = Field(default=None)
    connection_params: Optional[KafkaConnectionParamsPatch] = Field(default=None)
    # Full replacement only; users must provide complete TopicConfig entries
    topics: Optional[List[TopicConfig]] = Field(default=None)



from __future__ import annotations

from typing import Any, List

from . import errors, models
from .api_client import APIClient
from .pipeline import Pipeline


class Client(APIClient):
    """
    Manager class for handling multiple Pipeline instances.
    """

    ENDPOINT = "/api/v1/pipeline"

    def __init__(self, host: str | None = None) -> None:
        """Initialize the PipelineManager class.

        Args:
            host: GlassFlow API host
        """
        super().__init__(host=host)

    def get_pipeline(self, pipeline_id: str):
        """Fetch a pipeline by its ID.

        Args:
            pipeline_id: The ID of the pipeline to fetch

        Returns:
            Pipeline: A Pipeline instance for the given ID

        Raises:
            PipelineNotFoundError: If pipeline is not found
            APIError: If the API request fails
        """
        return Pipeline(host=self.host, pipeline_id=pipeline_id).get()

    def list_pipelines(self) -> List[dict]:
        """Returns a list of available pipelines.

        Returns:
            List[dict]: List of pipeline items with details as dictionaries

        Raises:
            APIError: If the API request fails
        """
        try:
            response = self._request("GET", self.ENDPOINT)
            data = response.json()

            # API always returns a list of pipelines
            return data if isinstance(data, list) else []

        except errors.NotFoundError:
            # No pipelines found, return empty list
            return []
        except errors.APIError as e:
            self._track_event("PipelineListError", error_type="InternalServerError")
            raise e

    def create_pipeline(
        self,
        pipeline_config: dict[str, Any] | models.PipelineConfig | None = None,
        pipeline_config_yaml_path: str | None = None,
        pipeline_config_json_path: str | None = None,
    ):
        """Creates a new pipeline with the given config.

        Args:
            pipeline_config: Dictionary or PipelineConfig object containing
                the pipeline configuration
            pipeline_config_yaml_path: Path to the YAML file containing
                the pipeline configuration
            pipeline_config_json_path: Path to the JSON file containing
                the pipeline configuration

        Returns:
            Pipeline: A Pipeline instance for the created pipeline

        Raises:
            PipelineAlreadyExistsError: If pipeline already exists
            PipelineInvalidConfigurationError: If configuration is invalid
            APIError: If the API request fails
        """
        if pipeline_config is None:
            if pipeline_config_yaml_path is None and pipeline_config_json_path is None:
                raise ValueError(
                    "Either pipeline_config or pipeline_config_yaml_path or "
                    "pipeline_config_json_path must be provided"
                )
            if pipeline_config_yaml_path is not None:
                pipeline = Pipeline.from_yaml(pipeline_config_yaml_path, host=self.host)
            elif pipeline_config_json_path is not None:
                pipeline = Pipeline.from_json(pipeline_config_json_path, host=self.host)
        else:
            if (
                pipeline_config_yaml_path is not None
                or pipeline_config_json_path is not None
            ):
                raise ValueError(
                    "Either pipeline_config or pipeline_config_yaml_path or "
                    "pipeline_config_json_path must be provided"
                )
            pipeline = Pipeline(config=pipeline_config, host=self.host)

        return pipeline.create()

    def stop_pipeline(self, pipeline_id: str, terminate: bool = False) -> None:
        """Stops the pipeline with the given ID.

        Args:
            pipeline_id: The ID of the pipeline to stop
            terminate: Whether to terminate the pipeline (i.e. delete all the pipeline
                components and potentially all the events in the pipeline)

        Raises:
            PipelineInTransitionError: If pipeline is in transition
            PipelineNotFoundError: If pipeline is not found
            APIError: If the API request fails
        """
        Pipeline(host=self.host, pipeline_id=pipeline_id).stop(terminate=terminate)

    def delete_pipeline(self, pipeline_id: str) -> None:
        """Deletes the pipeline with the given ID.

        Args:
            pipeline_id: The ID of the pipeline to delete

        Raises:
            PipelineDeletionStateViolationError: If pipeline is not stopped or
            terminating
            PipelineNotFoundError: If pipeline is not found
            APIError: If the API request fails
        """
        Pipeline(host=self.host, pipeline_id=pipeline_id).delete()

    def disable_tracking(self) -> None:
        """Disable tracking of pipeline events."""
        self._tracking.enabled = False



from __future__ import annotations

import os
import platform
from importlib.metadata import version
from typing import Any, Dict

import mixpanel


class Tracking:
    """Mixpanel tracking implementation for GlassFlow Clickhouse ETL."""

    def __init__(self, distinct_id: str) -> None:
        """Initialize the tracking client"""
        self.enabled = os.getenv("GF_TRACKING_ENABLED", "true").lower() == "true"
        self._project_token = "209670ec9b352915013a5dfdb169dd25"
        self._distinct_id = distinct_id
        self.client = mixpanel.Mixpanel(self._project_token)

        self.sdk_version = version("glassflow")
        self.platform = platform.system()
        self.python_version = platform.python_version()

    def track_event(
        self, event_name: str, properties: Dict[str, Any] | None = None
    ) -> None:
        """Track an event in Mixpanel.

        Args:
            event_name: Name of the event to track
            properties: Additional properties to include with the event
        """
        if not self.enabled:
            return

        base_properties = {
            "sdk_version": self.sdk_version,
            "platform": self.platform,
            "python_version": self.python_version,
        }
        if properties is None:
            properties = {}
        properties = {**base_properties, **properties}

        try:
            self.client.track(
                distinct_id=self._distinct_id,
                event_name=event_name,
                properties=properties,
            )
        except Exception:
            pass



from .base import CaseInsensitiveStrEnum


class KafkaDataType(CaseInsensitiveStrEnum):
    STRING = "string"
    INT = "int"
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"
    UINT = "uint"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    FLOAT = "float"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    BOOL = "bool"
    BYTES = "bytes"
    ARRAY = "array"


class ClickhouseDataType(CaseInsensitiveStrEnum):
    INT8 = "Int8"
    INT16 = "Int16"
    INT32 = "Int32"
    INT64 = "Int64"
    UINT8 = "UInt8"
    UINT16 = "UInt16"
    UINT32 = "UInt32"
    UINT64 = "UInt64"
    FLOAT32 = "Float32"
    FLOAT64 = "Float64"
    STRING = "String"
    FIXEDSTRING = "FixedString"
    DATETIME = "DateTime"
    DATETIME64 = "DateTime64"
    BOOL = "Bool"
    UUID = "UUID"
    ENUM8 = "Enum8"
    ENUM16 = "Enum16"
    LC_INT8 = "LowCardinality(Int8)"
    LC_INT16 = "LowCardinality(Int16)"
    LC_INT32 = "LowCardinality(Int32)"
    LC_INT64 = "LowCardinality(Int64)"
    LC_UINT8 = "LowCardinality(UInt8)"
    LC_UINT16 = "LowCardinality(UInt16)"
    LC_UINT32 = "LowCardinality(UInt32)"
    LC_UINT64 = "LowCardinality(UInt64)"
    LC_FLOAT32 = "LowCardinality(Float32)"
    LC_FLOAT64 = "LowCardinality(Float64)"
    LC_STRING = "LowCardinality(String)"
    LC_FIXEDSTRING = "LowCardinality(FixedString)"
    LC_DATETIME = "LowCardinality(DateTime)"
    ARRAY_STRING = "Array(String)"
    ARRAY_INT8 = "Array(Int8)"
    ARRAY_INT16 = "Array(Int16)"
    ARRAY_INT32 = "Array(Int32)"
    ARRAY_INT64 = "Array(Int64)"
    ARRAY_FLOAT32 = "Array(Float32)"
    ARRAY_FLOAT64 = "Array(Float64)"
    ARRAY_BOOL = "Array(Bool)"
    ARRAY_UINT8 = "Array(UInt8)"
    ARRAY_UINT16 = "Array(UInt16)"
    ARRAY_UINT32 = "Array(UInt32)"
    ARRAY_UINT64 = "Array(UInt64)"
    ARRAY_LC_STRING = "Array(LowCardinality(String))"
    ARRAY_LC_INT8 = "Array(LowCardinality(Int8))"
    ARRAY_LC_INT16 = "Array(LowCardinality(Int16))"
    ARRAY_LC_INT32 = "Array(LowCardinality(Int32))"
    ARRAY_LC_INT64 = "Array(LowCardinality(Int64))"
    ARRAY_LC_UINT8 = "Array(LowCardinality(UInt8))"
    ARRAY_LC_UINT16 = "Array(LowCardinality(UInt16))"
    ARRAY_LC_UINT32 = "Array(LowCardinality(UInt32))"
    ARRAY_LC_UINT64 = "Array(LowCardinality(UInt64))"
    ARRAY_LC_FLOAT32 = "Array(LowCardinality(Float32))"
    ARRAY_LC_FLOAT64 = "Array(LowCardinality(Float64))"
    ARRAY_LC_DATETIME = "Array(LowCardinality(DateTime))"
    ARRAY_LC_FIXEDSTRING = "Array(LowCardinality(FixedString))"


kafka_to_clickhouse_data_type_mappings = {
    KafkaDataType.STRING: [
        ClickhouseDataType.STRING,
        ClickhouseDataType.FIXEDSTRING,
        ClickhouseDataType.DATETIME,
        ClickhouseDataType.DATETIME64,
        ClickhouseDataType.UUID,
        ClickhouseDataType.ENUM8,
        ClickhouseDataType.ENUM16,
        ClickhouseDataType.LC_STRING,
        ClickhouseDataType.LC_FIXEDSTRING,
        ClickhouseDataType.LC_DATETIME,
    ],
    KafkaDataType.INT: [
        ClickhouseDataType.INT8,
        ClickhouseDataType.INT16,
        ClickhouseDataType.INT32,
        ClickhouseDataType.INT64,
    ],
    KafkaDataType.INT8: [ClickhouseDataType.INT8, ClickhouseDataType.LC_INT8],
    KafkaDataType.INT16: [ClickhouseDataType.INT16, ClickhouseDataType.LC_INT16],
    KafkaDataType.INT32: [ClickhouseDataType.INT32, ClickhouseDataType.LC_INT32],
    KafkaDataType.INT64: [
        ClickhouseDataType.INT64,
        ClickhouseDataType.DATETIME,
        ClickhouseDataType.DATETIME64,
        ClickhouseDataType.LC_INT64,
        ClickhouseDataType.LC_DATETIME,
    ],
    KafkaDataType.UINT: [
        ClickhouseDataType.UINT8,
        ClickhouseDataType.UINT16,
        ClickhouseDataType.UINT32,
        ClickhouseDataType.UINT64,
    ],
    KafkaDataType.UINT8: [ClickhouseDataType.UINT8],
    KafkaDataType.UINT16: [ClickhouseDataType.UINT16],
    KafkaDataType.UINT32: [ClickhouseDataType.UINT32],
    KafkaDataType.UINT64: [ClickhouseDataType.UINT64],
    KafkaDataType.UINT8: [ClickhouseDataType.UINT8],
    KafkaDataType.FLOAT: [ClickhouseDataType.FLOAT32, ClickhouseDataType.FLOAT64],
    KafkaDataType.FLOAT32: [ClickhouseDataType.FLOAT32, ClickhouseDataType.LC_FLOAT32],
    KafkaDataType.FLOAT64: [
        ClickhouseDataType.FLOAT64,
        ClickhouseDataType.DATETIME,
        ClickhouseDataType.DATETIME64,
        ClickhouseDataType.LC_FLOAT64,
        ClickhouseDataType.LC_DATETIME,
    ],
    KafkaDataType.BOOL: [ClickhouseDataType.BOOL],
    KafkaDataType.BYTES: [ClickhouseDataType.STRING],
    KafkaDataType.ARRAY: [
        ClickhouseDataType.STRING,
        ClickhouseDataType.ARRAY_STRING,
        ClickhouseDataType.ARRAY_INT8,
        ClickhouseDataType.ARRAY_INT16,
        ClickhouseDataType.ARRAY_INT32,
        ClickhouseDataType.ARRAY_INT64,
        ClickhouseDataType.ARRAY_FLOAT32,
        ClickhouseDataType.ARRAY_FLOAT64,
        ClickhouseDataType.ARRAY_BOOL,
        ClickhouseDataType.ARRAY_UINT8,
        ClickhouseDataType.ARRAY_UINT16,
        ClickhouseDataType.ARRAY_UINT32,
        ClickhouseDataType.ARRAY_UINT64,
        ClickhouseDataType.ARRAY_LC_STRING,
        ClickhouseDataType.ARRAY_LC_INT8,
        ClickhouseDataType.ARRAY_LC_INT16,
        ClickhouseDataType.ARRAY_LC_INT32,
        ClickhouseDataType.ARRAY_LC_INT64,
        ClickhouseDataType.ARRAY_LC_UINT8,
        ClickhouseDataType.ARRAY_LC_UINT16,
        ClickhouseDataType.ARRAY_LC_UINT32,
        ClickhouseDataType.ARRAY_LC_UINT64,
        ClickhouseDataType.ARRAY_LC_FLOAT32,
        ClickhouseDataType.ARRAY_LC_FLOAT64,
        ClickhouseDataType.ARRAY_LC_DATETIME,
        ClickhouseDataType.ARRAY_LC_FIXEDSTRING,
    ],
}



import re
from typing import Any, Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from ..errors import InvalidDataTypeMappingError
from .base import CaseInsensitiveStrEnum
from .data_types import kafka_to_clickhouse_data_type_mappings
from .join import JoinConfig, JoinConfigPatch
from .sink import SinkConfig, SinkConfigPatch
from .source import SourceConfig, SourceConfigPatch


class PipelineStatus(CaseInsensitiveStrEnum):
    CREATED = "Created"
    RUNNING = "Running"
    RESUMING = "Resuming"
    STOPPING = "Stopping"
    STOPPED = "Stopped"
    TERMINATING = "Terminating"
    TERMINATED = "Terminated"
    FAILED = "Failed"
    DELETED = "Deleted"


class PipelineConfig(BaseModel):
    pipeline_id: str
    name: Optional[str] = Field(default=None)
    source: SourceConfig
    join: Optional[JoinConfig] = Field(default=JoinConfig())
    sink: SinkConfig

    @field_validator("pipeline_id")
    @classmethod
    def validate_pipeline_id(cls, v: str) -> str:
        if not v:
            raise ValueError("pipeline_id cannot be empty")
        if len(v) > 40:
            raise ValueError("pipeline_id cannot be longer than 40 characters")
        if not re.match(r"^[a-z0-9-]+$", v):
            raise ValueError(
                "pipeline_id can only contain lowercase letters, numbers, and hyphens"
            )
        if not re.match(r"^[a-z0-9]", v):
            raise ValueError("pipeline_id must start with a lowercase alphanumeric")
        if not re.match(r".*[a-z0-9]$", v):
            raise ValueError("pipeline_id must end with a lowercase alphanumeric")
        return v

    @model_validator(mode="after")
    def set_pipeline_name(self) -> "PipelineConfig":
        """
        If name is not provided, use the pipeline_id and replace hyphens
        with spaces.
        """
        if self.name is None:
            self.name = self.pipeline_id.replace("-", " ").title()
        return self

    @field_validator("join")
    @classmethod
    def validate_join_config(
        cls,
        v: Optional[JoinConfig],
        info: Any,
    ) -> Optional[JoinConfig]:
        if not v or not v.enabled:
            return v

        # Get the source topics from the parent model's data
        source = info.data.get("source", {})
        if isinstance(source, dict):
            source_topics = source.get("topics", [])
        else:
            source_topics = source.topics
        if not source_topics:
            return v

        # Validate each source in the join config
        for source in v.sources:
            # Check if source_id exists in any topic
            source_exists = any(
                topic.name == source.source_id for topic in source_topics
            )
            if not source_exists:
                raise ValueError(
                    f"Source ID '{source.source_id}' does not exist in any topic"
                )

            # Find the topic and check if join_key exists in its schema
            topic = next((t for t in source_topics if t.name == source.source_id), None)
            if not topic:
                continue

            field_exists = any(
                field.name == source.join_key for field in topic.event_schema.fields
            )
            if not field_exists:
                raise ValueError(
                    f"Join key '{source.join_key}' does not exist in source "
                    f"'{source.source_id}' schema"
                )

        return v

    @field_validator("sink")
    @classmethod
    def validate_sink_config(cls, v: SinkConfig, info: Any) -> SinkConfig:
        # Get the source topics from the parent model's data
        source = info.data.get("source", {})
        if isinstance(source, dict):
            source_topics = source.get("topics", [])
        else:
            source_topics = source.topics
        if not source_topics:
            return v

        # Validate each table mapping
        for mapping in v.table_mapping:
            # Check if source_id exists in any topic
            source_exists = any(
                topic.name == mapping.source_id for topic in source_topics
            )
            if not source_exists:
                raise ValueError(
                    f"Source ID '{mapping.source_id}' does not exist in any topic"
                )

            # Find the topic and check if field_name exists in its schema
            topic = next(
                (t for t in source_topics if t.name == mapping.source_id), None
            )
            if not topic:
                continue

            field_exists = any(
                field.name == mapping.field_name for field in topic.event_schema.fields
            )
            if not field_exists:
                raise ValueError(
                    f"Field '{mapping.field_name}' does not exist in source "
                    f"'{mapping.source_id}' event schema"
                )

        return v

    @field_validator("sink")
    @classmethod
    def validate_data_type_compatibility(cls, v: SinkConfig, info: Any) -> SinkConfig:
        # Get the source topics from the parent model's data
        source = info.data.get("source", {})
        if isinstance(source, dict):
            source_topics = source.get("topics", [])
        else:
            source_topics = source.topics
        if not source_topics:
            return v

        # Validate each table mapping
        for mapping in v.table_mapping:
            # Find the topic
            topic = next(
                (t for t in source_topics if t.name == mapping.source_id), None
            )
            if not topic:
                continue

            # Find the source field
            source_field = next(
                (f for f in topic.event_schema.fields if f.name == mapping.field_name),
                None,
            )
            if not source_field:
                continue

            # Get the source and target data types
            source_type = source_field.type
            target_type = mapping.column_type

            # Check if the target type is compatible with the source type
            compatible_types = kafka_to_clickhouse_data_type_mappings.get(
                source_type, []
            )
            if target_type not in compatible_types:
                raise InvalidDataTypeMappingError(
                    f"Data type '{target_type}' is not compatible with source type "
                    f"'{source_type}' for field '{mapping.field_name}' in source "
                    f"'{mapping.source_id}'"
                )

        return v

    def update(self, config_patch: "PipelineConfigPatch") -> "PipelineConfig":
        """
        Apply a patch configuration to this pipeline configuration.

        Args:
            config_patch: The patch configuration (PipelineConfigPatch or dict)

        Returns:
            PipelineConfig: A new PipelineConfig instance with the patch applied
        """
        # Start with a deep copy of the current config
        updated_config = self.model_copy(deep=True)

        # Update name if provided
        if config_patch.name is not None:
            updated_config.name = config_patch.name

        # Update source if provided
        if config_patch.source is not None:
            updated_config.source = updated_config.source.update(config_patch.source)

        # Update join if provided
        if config_patch.join is not None:
            updated_config.join = (updated_config.join or JoinConfig()).update(
                config_patch.join
            )

        # Update sink if provided
        if config_patch.sink is not None:
            updated_config.sink = updated_config.sink.update(config_patch.sink)

        return updated_config


class PipelineConfigPatch(BaseModel):
    name: Optional[str] = Field(default=None)
    source: Optional[SourceConfigPatch] = Field(default=None)
    join: Optional[JoinConfigPatch] = Field(default=None)
    sink: Optional[SinkConfigPatch] = Field(default=None)



import configparser
import os
import uuid
from pathlib import Path

from pydantic import BaseModel


class GlassFlowSettings(BaseModel):
    host: str = "http://localhost:8080"


class AnalyticsSettings(BaseModel):
    enabled: bool = True
    distinct_id: str = str(uuid.uuid4())


class GlassFlowConfig(BaseModel):
    config_file: Path = Path.home() / ".glassflow" / "clickhouse.conf"
    glassflow: GlassFlowSettings = GlassFlowSettings()
    analytics: AnalyticsSettings = AnalyticsSettings()

    def model_post_init(self, ctx):
        self.load()

    def load(self):
        config = configparser.ConfigParser()

        if os.path.exists(self.config_file):
            with open(self.config_file, "r") as f:
                config.read_file(f)
            if "glassflow" in config:
                self.glassflow = GlassFlowSettings(**config["glassflow"])
            if "analytics" in config:
                self.analytics = AnalyticsSettings(**config["analytics"])
        else:
            self.write()

    def write(self):
        if not os.path.exists(self.config_file.parent):
            os.makedirs(self.config_file.parent)

        config = configparser.ConfigParser()
        config["glassflow"] = self.glassflow.model_dump()
        config["analytics"] = self.analytics.model_dump()

        with open(self.config_file, "w") as f:
            config.write(f)



from __future__ import annotations

import json
from typing import Any

import httpx

from . import errors
from .models import GlassFlowConfig
from .tracking import Tracking


class APIClient:
    """
    API client
    """

    version = "1"
    glassflow_config = GlassFlowConfig()
    _tracking = Tracking(glassflow_config.analytics.distinct_id)

    def __init__(self, host: str | None = None):
        """Initialize the API Client class.

        Args:
            host: Host URL of the GlassFlow Clickhouse ETL service
        """
        self.host = host if host else self.glassflow_config.glassflow.host
        self.http_client = httpx.Client(base_url=self.host)

    def _request(
        self, method: str, endpoint: str, **kwargs: Any
    ) -> httpx.Response | None:
        """
        Generic request method with centralized error handling.

        Args:
            method: HTTP method (GET, POST, DELETE, etc.)
            endpoint: API endpoint
            **kwargs: Additional arguments to pass to httpx

        Returns:
            httpx.Response: The response object

        Raises:
            httpx.HTTPStatusError: If the API request fails with HTTP errors
                (to be handled by subclasses)
            RequestError: If there is a network error
        """
        try:
            response = self.http_client.request(method, endpoint, **kwargs)
            response.raise_for_status()
            return response
        except httpx.HTTPStatusError as e:
            self._raise_api_error(e.response)
        except httpx.RequestError as e:
            self._track_event("RequestError", error_type="ConnectionError")
            raise errors.ConnectionError(
                "Failed to connect to GlassFlow ETL API"
            ) from e

    @staticmethod
    def _raise_api_error(response: httpx.Response) -> None:
        """Raise an APIError based on the response."""
        status_code = response.status_code
        try:
            error_data = response.json()
            message = error_data.get("message", None)
            code = error_data.get("code", None)
        except json.JSONDecodeError:
            message = f"{status_code} {response.reason_phrase}"
            code = None
            error_data = {}

        if status_code == 400:
            # Handle specific status validation error codes
            if code == "TERMINAL_STATE_VIOLATION":
                raise errors.TerminalStateViolationError(
                    status_code, message, response=response
                )
            elif code == "INVALID_STATUS_TRANSITION":
                raise errors.InvalidStatusTransitionError(
                    status_code,
                    message,
                    response=response,
                )
            elif code == "UNKNOWN_STATUS":
                raise errors.UnknownStatusError(status_code, message, response=response)
            elif code == "PIPELINE_ALREADY_IN_STATE":
                raise errors.PipelineAlreadyInStateError(
                    status_code, message, response=response
                )
            elif code == "PIPELINE_IN_TRANSITION":
                raise errors.PipelineInTransitionError(
                    status_code, message, response=response
                )
            elif message and message.startswith("invalid json:"):
                raise errors.InvalidJsonError(status_code, message, response=response)
            elif message and message == "pipeline id cannot be empty":
                raise errors.EmptyPipelineIdError(
                    status_code, message, response=response
                )
            elif message and message.startswith(
                "pipeline can only be deleted if it's stopped or terminated"
            ):
                raise errors.PipelineDeletionStateViolationError(
                    status_code, message, response=response
                )
            else:
                # Generic 400 error for unknown codes
                raise errors.ValidationError(status_code, message, response=response)
        elif status_code == 403:
            raise errors.ForbiddenError(status_code, message, response=response)
        elif status_code == 404:
            raise errors.NotFoundError(status_code, message, response=response)
        elif status_code == 422:
            raise errors.UnprocessableContentError(
                status_code, message, response=response
            )
        elif status_code == 500:
            raise errors.ServerError(status_code, message, response=response)
        else:
            raise errors.APIError(
                status_code,
                message="An error occurred: "
                f"({status_code} {response.reason_phrase}) {message}",
                response=response,
            )

    def _track_event(self, event_name: str, **kwargs: Any) -> None:
        """Track an event with the given name and properties."""
        self._tracking.track_event(event_name, kwargs)



"""
GlassFlow SDK for creating data pipelines between Kafka and ClickHouse.
"""

from .client import Client
from .dlq import DLQ
from .models import (
    JoinConfig,
    PipelineConfig,
    SinkConfig,
    SourceConfig,
)
from .pipeline import Pipeline

__all__ = [
    "Pipeline",
    "Client",
    "DLQ",
    "PipelineConfig",
    "SourceConfig",
    "SinkConfig",
    "JoinConfig",
]



class GlassFlowError(Exception):
    """Base exception for all GlassFlow SDK errors."""


# Generic client-side errors
class RequestError(GlassFlowError):
    """A low-level HTTP or network error."""


class ConnectionError(RequestError):
    """Raised when a connection to the server fails."""


# Server/API-level errors
class APIError(GlassFlowError):
    """Base for API response errors."""

    def __init__(self, status_code, message=None, response=None):
        self.status_code = status_code
        self.response = response
        self.message = message
        super().__init__(self.message)


class NotFoundError(APIError):
    """Raised when a resource is not found (404)."""


class ValidationError(APIError):
    """Raised on 400 Bad Request errors due to bad input."""


class ForbiddenError(APIError):
    """Raised on 403 Forbidden errors."""


class UnprocessableContentError(APIError):
    """Raised on 422 Unprocessable Content errors."""


class ServerError(APIError):
    """Raised on 500 Server Error errors."""


class PipelineAlreadyExistsError(APIError):
    """Raised when a pipeline already exists."""


class PipelineNotFoundError(APIError):
    """Raised on 404 when a pipeline is not found."""


class PipelineInvalidConfigurationError(APIError):
    """Raised when a pipeline configuration is invalid."""


class InvalidDataTypeMappingError(GlassFlowError):
    """Exception raised when a data type mapping is invalid."""


class InvalidBatchSizeError(GlassFlowError):
    """Exception raised when a batch size is invalid."""


# Status validation error classes for 400 Bad Request responses
class TerminalStateViolationError(ValidationError):
    """Raised when attempting to transition from a terminal state to another state."""


class InvalidStatusTransitionError(ValidationError):
    """Raised when attempting an invalid status transition."""


class UnknownStatusError(ValidationError):
    """Raised when an unknown pipeline status is encountered."""


class PipelineAlreadyInStateError(ValidationError):
    """Raised when pipeline is already in the requested state."""


class PipelineInTransitionError(ValidationError):
    """
    Raised when pipeline is currently transitioning and cannot perform the
    requested operation.
    """


class InvalidJsonError(ValidationError):
    """Raised when malformed JSON is provided in request body."""


class EmptyPipelineIdError(ValidationError):
    """Raised when pipeline ID parameter is empty or whitespace."""


class PipelineDeletionStateViolationError(ValidationError):
    """Raised when attempting to delete a pipeline that's not in a deletable state."""



from typing import Any, List, Optional

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from .base import CaseInsensitiveStrEnum


class JoinOrientation(CaseInsensitiveStrEnum):
    LEFT = "left"
    RIGHT = "right"


class JoinSourceConfig(BaseModel):
    source_id: str
    join_key: str
    time_window: str
    orientation: JoinOrientation


class JoinType(CaseInsensitiveStrEnum):
    TEMPORAL = "temporal"


class JoinConfig(BaseModel):
    """Configuration for joining multiple sources."""

    enabled: bool = False
    type: Optional[JoinType] = None
    sources: Optional[List[JoinSourceConfig]] = None

    @field_validator("sources")
    @classmethod
    def validate_sources(
        cls, v: Optional[List[JoinSourceConfig]], info: ValidationInfo
    ) -> Optional[List[JoinSourceConfig]]:
        """
        Validate that when join is enabled, there are exactly two sources
        with opposite orientations.
        """
        if not info.data.get("enabled", False):
            return v

        if not v:
            raise ValueError("sources are required when join is enabled")

        if len(v) != 2:
            raise ValueError("join must have exactly two sources when enabled")

        orientations = {source.orientation for source in v}
        if orientations != {JoinOrientation.LEFT, JoinOrientation.RIGHT}:
            raise ValueError(
                "join sources must have opposite orientations (one LEFT and one RIGHT)"
            )

        return v

    @field_validator("type")
    @classmethod
    def validate_type(
        cls, v: Optional[JoinType], info: ValidationInfo
    ) -> Optional[JoinType]:
        """Validate that type is required when join is enabled."""
        if info.data.get("enabled", False) and not v:
            raise ValueError("type is required when join is enabled")
        return v

    def update(self, patch: "JoinConfigPatch") -> "JoinConfig":
        """Apply a patch to this join config."""
        update_dict: dict[str, Any] = {}

        if patch.enabled is not None:
            update_dict["enabled"] = patch.enabled
        if patch.type is not None:
            update_dict["type"] = patch.type
        if patch.sources is not None:
            update_dict["sources"] = patch.sources

        if update_dict:
            return self.model_copy(update=update_dict)
        return self


class JoinConfigPatch(BaseModel):
    enabled: Optional[bool] = Field(default=None)
    type: Optional[JoinType] = Field(default=None)
    sources: Optional[List[JoinSourceConfig]] = Field(default=None)



from __future__ import annotations

import json
from typing import Any

import yaml
from httpx._models import Response
from pydantic import ValidationError

from . import errors, models
from .api_client import APIClient
from .dlq import DLQ


class Pipeline(APIClient):
    """
    Main class for managing Kafka to ClickHouse pipelines.
    """

    ENDPOINT = "/api/v1/pipeline"

    def __init__(
        self,
        host: str | None = None,
        pipeline_id: str | None = None,
        config: models.PipelineConfig | dict[str, Any] | None = None,
    ):
        """Initialize the Pipeline class.

        Args:
            host: GlassFlow API host
            pipeline_id: ID of the pipeline to create
            config: Pipeline configuration
        """
        super().__init__(host=host)

        if not config and not pipeline_id:
            raise ValueError("Either config or pipeline_id must be provided")
        elif config and pipeline_id:
            raise ValueError("Only one of config or pipeline_id can be provided")

        if pipeline_id is not None:
            self.pipeline_id = pipeline_id

        if config is not None:
            if isinstance(config, dict):
                self.config = models.PipelineConfig.model_validate(config)
            else:
                self.config = config
            self.pipeline_id = self.config.pipeline_id
        else:
            self.config = None

        self._dlq = DLQ(pipeline_id=self.pipeline_id, host=host)
        self.status: models.PipelineStatus | None = None

    def get(self) -> Pipeline:
        """Fetch a pipeline by its ID.

        Returns:
            Pipeline: A Pipeline instance for the given ID

        Raises:
            PipelineNotFoundError: If pipeline is not found
            APIError: If the API request fails
        """
        response = self._request(
            "GET", f"{self.ENDPOINT}/{self.pipeline_id}", event_name="PipelineGet"
        )
        self.config = models.PipelineConfig.model_validate(response.json())
        self.health()
        self._dlq = DLQ(pipeline_id=self.pipeline_id, host=self.host)
        return self

    def create(self) -> Pipeline:
        """Creates a new pipeline with the given config.

        Returns:
            Pipeline: A Pipeline instance for the created pipeline

        Raises:
            PipelineAlreadyExistsError: If pipeline already exists
            PipelineInvalidConfigurationError: If configuration is invalid
            APIError: If the API request fails
        """
        if self.config is None:
            raise ValueError("Pipeline configuration must be provided in constructor")
        try:
            self._request(
                "POST",
                self.ENDPOINT,
                json=self.config.model_dump(
                    mode="json",
                    by_alias=True,
                    exclude_none=True,
                ),
                event_name="PipelineCreated",
            )
            self.status = models.PipelineStatus.CREATED
            return self

        except errors.ForbiddenError as e:
            self._track_event("PipelineCreated", error_type="PipelineAlreadyExists")
            raise errors.PipelineAlreadyExistsError(
                status_code=e.status_code,
                message=f"Pipeline with ID {self.config.pipeline_id} already exists;"
                "delete it first before creating new pipeline or use a"
                "different pipeline ID",
                response=e.response,
            ) from e

    def rename(self, name: str) -> Pipeline:
        """Renames the pipeline with the given name.

        Returns:
            Pipeline: A Pipeline instance for the renamed pipeline

        Raises:
            PipelineNotFoundError: If pipeline is not found
            APIError: If the API request fails
        """
        self._request(
            "PATCH",
            f"{self.ENDPOINT}/{self.pipeline_id}",
            json={"name": name},
            event_name="PipelineRenamed",
        )
        self.config.name = name
        return self

    def update(
        self, config_patch: models.PipelineConfigPatch | dict[str, Any]
    ) -> Pipeline:
        """Updates the pipeline with the given config patch.
        Pipeline must be stopped or terminated before updating.

        Args:
            config_patch: Pipeline configuration patch

        Returns:
            Pipeline: A Pipeline instance for the updated pipeline

        Raises:
            PipelineNotFoundError: If pipeline is not found
            PipelineInTransitionError: If pipeline is in transition
            InvalidStatusTransitionError: If pipeline is not in a state that can be
                updated
            APIError: If the API request fails
        """
        self.get()  # Get latest config
        if isinstance(config_patch, dict):
            config_patch = models.PipelineConfigPatch.model_validate(config_patch)
        else:
            config_patch = config_patch
        updated_config = self.config.update(config_patch)

        self._request(
            "POST",
            f"{self.ENDPOINT}/{self.pipeline_id}/edit",
            json=updated_config.model_dump(
                mode="json",
                by_alias=True,
                exclude_none=True,
            ),
            event_name="PipelineUpdated",
        )
        self.status = models.PipelineStatus.RESUMING

        # Update self.config with the updated configuration
        self.config = updated_config
        return self

    def delete(self) -> None:
        """
        Deletes the pipeline from the database. Only pipelines that are stopped or
        terminated can be deleted.

        Raises:
            PipelineDeletionStateViolationError: If pipeline is not stopped or
                terminated
            PipelineNotFoundError: If pipeline is not found
            APIError: If the API request fails
        """
        endpoint = f"{self.ENDPOINT}/{self.pipeline_id}"
        self._request("DELETE", endpoint, event_name="PipelineDeleted")
        self.status = models.PipelineStatus.DELETED

    def stop(self, terminate: bool = False) -> Pipeline:
        """
        Stops the pipeline, waiting for all the events in the pipeline to be processed.
        If terminate is True, the pipeline will be terminated instead.

        Args:
            terminate: Whether to terminate the pipeline (i.e. delete all the pipeline
                components and potentially all the events in the pipeline)

        Returns:
            Pipeline: A Pipeline instance for the stopped pipeline

        Raises:
            PipelineInTransitionError: If pipeline is in transition
            PipelineNotFoundError: If pipeline is not found
            InvalidStatusTransitionError: If pipeline is not in a state that can be
                stopped
            APIError: If the API request fails
        """
        if terminate:
            endpoint = f"{self.ENDPOINT}/{self.pipeline_id}/terminate"
            next_status = models.PipelineStatus.TERMINATING
            event_name = "PipelineTerminated"
        else:
            endpoint = f"{self.ENDPOINT}/{self.pipeline_id}/stop"
            next_status = models.PipelineStatus.STOPPING
            event_name = "PipelineStopped"
        self._request("POST", endpoint, event_name=event_name)
        self.status = next_status
        return self

    def resume(self) -> Pipeline:
        """
        Resumes the pipeline with the given ID.
        Only stopped or terminated pipelines can be resumed.

        Returns:
            Pipeline: A Pipeline instance for the resumed pipeline

        Raises:
            PipelineInTransitionError: If pipeline is in transition
            PipelineNotFoundError: If pipeline is not found
            InvalidStatusTransitionError: If pipeline is not in a state that can be
                resumed
            APIError: If the API request fails
        """
        endpoint = f"{self.ENDPOINT}/{self.pipeline_id}/resume"
        self._request("POST", endpoint, event_name="PipelineResumed")
        self.status = models.PipelineStatus.RESUMING
        return self

    def health(self) -> dict[str, Any]:
        """Get the health of the pipeline.

        Returns:
            dict: Pipeline health
        """
        response = self._request(
            "GET",
            f"{self.ENDPOINT}/{self.pipeline_id}/health",
            event_name="PipelineHealth",
        ).json()
        self.status = models.PipelineStatus(response["overall_status"])
        return response

    def to_dict(self) -> dict[str, Any]:
        """Convert the pipeline configuration to a dictionary.

        Returns:
            dict: Pipeline configuration as a dictionary
        """
        if not hasattr(self, "config") or self.config is None:
            return {"pipeline_id": self.pipeline_id}

        return self.config.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
        )

    def to_yaml(self, yaml_path: str) -> None:
        """Save the pipeline configuration to a YAML file.

        Args:
            yaml_path: Path to the YAML file
        """
        with open(yaml_path, "w") as f:
            yaml.dump(self.to_dict(), f, default_flow_style=False)

    def to_json(self, json_path: str) -> None:
        """Save the pipeline configuration to a JSON file.

        Args:
            json_path: Path to the JSON file
        """
        with open(json_path, "w") as f:
            json.dump(self.to_dict(), f, indent=4)

    @classmethod
    def from_yaml(cls, yaml_path: str, host: str | None = None) -> Pipeline:
        """Create a pipeline from a YAML file.

        Args:
            yaml_path: Path to the YAML file
            host: GlassFlow API host

        Returns:
            Pipeline: A Pipeline instance for the created pipeline
        """
        with open(yaml_path, "r") as f:
            config = yaml.safe_load(f)
        return cls(config=config, host=host)

    @classmethod
    def from_json(cls, json_path: str, host: str | None = None) -> Pipeline:
        """Create a pipeline from a JSON file.

        Args:
            json_path: Path to the JSON file
            host: GlassFlow API host

        Returns:
            Pipeline: A Pipeline instance for the created pipeline
        """
        with open(json_path, "r") as f:
            config = json.load(f)
        return cls(config=config, host=host)

    @staticmethod
    def validate_config(config: dict[str, Any]) -> bool:
        """
        Validate a pipeline configuration.

        Args:
            config: Dictionary containing the pipeline configuration

        Returns:
            True if the configuration is valid

        Raises:
            ValueError: If the configuration is invalid
            ValidationError: If the configuration fails Pydantic validation
        """
        try:
            models.PipelineConfig.model_validate(config)
            return True
        except ValidationError as e:
            raise e
        except ValueError as e:
            raise e

    @property
    def dlq(self) -> DLQ:
        """Get the DLQ (Dead Letter Queue) client for this pipeline.

        Returns:
            DLQ: The DLQ client instance
        """
        return self._dlq

    @dlq.setter
    def dlq(self, dlq: DLQ) -> None:
        self._dlq = dlq

    def _tracking_info(self) -> dict[str, Any]:
        """Get information about the active pipeline."""
        # If config is not set, return minimal info
        if self.config is None:
            return {"pipeline_id": self.pipeline_id}

        # Extract join info
        join_enabled = getattr(self.config.join, "enabled", False)

        # Extract deduplication info
        deduplication_enabled = any(
            t.deduplication and t.deduplication.enabled
            for t in self.config.source.topics
        )

        # Extract connection params
        conn_params = self.config.source.connection_params

        root_ca_provided = conn_params.root_ca is not None
        skip_auth = conn_params.skip_auth
        protocol = str(conn_params.protocol)
        mechanism = str(conn_params.mechanism)

        return {
            "pipeline_id": self.config.pipeline_id,
            "join_enabled": join_enabled,
            "deduplication_enabled": deduplication_enabled,
            "source_auth_method": mechanism,
            "source_security_protocol": protocol,
            "source_root_ca_provided": root_ca_provided,
            "source_skip_auth": skip_auth,
        }

    def _track_event(self, event_name: str, **kwargs: Any) -> None:
        pipeline_properties = self._tracking_info()
        properties = {**pipeline_properties, **kwargs}
        super()._track_event(event_name, **properties)

    def _request(
        self, method: str, endpoint: str, event_name: str, **kwargs: Any
    ) -> Response:
        try:
            response = super()._request(method, endpoint, **kwargs)
            self._track_event(event_name)
            return response
        except errors.NotFoundError as e:
            self._track_event(event_name, error_type="PipelineNotFound")
            raise errors.PipelineNotFoundError(
                status_code=e.status_code,
                message=f"Pipeline with id '{self.pipeline_id}' not found",
                response=e.response,
            ) from e
        except errors.UnprocessableContentError as e:
            self._track_event(event_name, error_type="InvalidPipelineConfig")
            raise errors.PipelineInvalidConfigurationError(
                status_code=e.status_code,
                message=e.message or "Invalid pipeline configuration",
            ) from e
        except errors.APIError as e:
            self._track_event(event_name, error_type="InternalServerError")
            raise e



from __future__ import annotations

from typing import Any, Dict, List

from . import errors
from .api_client import APIClient
from .errors import InvalidBatchSizeError


class DLQ(APIClient):
    """
    Dead Letter Queue client for managing failed messages.
    """

    def __init__(self, pipeline_id: str, host: str | None = None):
        super().__init__(host)
        self.pipeline_id = pipeline_id
        self.endpoint = f"/api/v1/pipeline/{self.pipeline_id}/dlq"
        self._max_batch_size = 100

    def consume(self, batch_size: int = 100) -> List[Dict[str, Any]]:
        """
        Consume messages from the Dead Letter Queue.

        Args:
            batch_size: Number of messages to consume (between 1 and 1000)

        Returns:
            List of messages from the DLQ
        """
        if (
            not isinstance(batch_size, int)
            or batch_size < 1
            or batch_size > self._max_batch_size
        ):
            raise ValueError("batch_size must be an integer between 1 and 100")

        try:
            response = self._request(
                "GET", f"{self.endpoint}/consume", params={"batch_size": batch_size}
            )
            response.raise_for_status()
            if response.status_code != 204:
                return response.json()
            return []
        except errors.UnprocessableContentError as e:
            raise InvalidBatchSizeError(
                f"Invalid batch size: batch size should be larger than 1 "
                f"and smaller than {self._max_batch_size}"
            ) from e
        except errors.APIError as e:
            raise e

    def state(self) -> Dict[str, Any]:
        """
        Get the current state of the Dead Letter Queue.

        Returns:
            Dictionary containing DLQ state information

        Raises:
            ConnectionError: If there is a network error
            InternalServerError: If the API request fails
        """
        try:
            response = self._request("GET", f"{self.endpoint}/state")
            response.raise_for_status()
            return response.json()
        except errors.NotFoundError as e:
            raise errors.PipelineNotFoundError(
                status_code=e.status_code,
                message=f"Pipeline with id '{self.pipeline_id}' not found",
                response=e.response,
            ) from e
        except errors.APIError as e:
            raise e

    def purge(self) -> None:
        """
        Purge all messages from the Dead Letter Queue.

        This operation removes all messages currently in the DLQ and cannot be undone.

        Raises:
            PipelineNotFoundError: If the pipeline does not exist
            ConnectionError: If there is a network error
            InternalServerError: If the API request fails
        """
        try:
            response = self._request("POST", f"{self.endpoint}/purge")
            response.raise_for_status()
        except errors.NotFoundError as e:
            raise errors.PipelineNotFoundError(
                status_code=e.status_code,
                message=f"Pipeline with id '{self.pipeline_id}' not found",
                response=e.response,
            ) from e
        except errors.APIError as e:
            raise e



from .config import GlassFlowConfig
from .data_types import ClickhouseDataType, KafkaDataType
from .join import (
    JoinConfig,
    JoinConfigPatch,
    JoinOrientation,
    JoinSourceConfig,
    JoinType,
)
from .pipeline import PipelineConfig, PipelineConfigPatch, PipelineStatus
from .sink import SinkConfig, SinkConfigPatch, SinkType, TableMapping
from .source import (
    ConsumerGroupOffset,
    DeduplicationConfig,
    DeduplicationConfigPatch,
    KafkaConnectionParams,
    KafkaConnectionParamsPatch,
    KafkaMechanism,
    Schema,
    SchemaField,
    SchemaType,
    SourceConfig,
    SourceConfigPatch,
    SourceType,
    TopicConfig,
)

__all__ = [
    "ClickhouseDataType",
    "ConsumerGroupOffset",
    "DeduplicationConfig",
    "KafkaConnectionParams",
    "KafkaDataType",
    "KafkaMechanism",
    "JoinConfig",
    "JoinOrientation",
    "JoinSourceConfig",
    "JoinType",
    "PipelineConfig",
    "PipelineConfigPatch",
    "PipelineStatus",
    "SinkConfig",
    "SinkConfigPatch",
    "SinkType",
    "TableMapping",
    "Schema",
    "SchemaField",
    "SchemaType",
    "SourceConfig",
    "SourceType",
    "TopicConfig",
    "GlassFlowConfig",
    "SourceConfigPatch",
    "KafkaConnectionParamsPatch",
    "DeduplicationConfigPatch",
    "JoinConfigPatch",
]
